/*
 * Copyright (C) 2009 Red Hat <nfs@redhat.com>
 *
 * support/export/v4root.c
 *
 * Routines used to support NFSv4 pseudo roots
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/queue.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <unistd.h>
#include <errno.h>

#include "xlog.h"
#include "exportfs.h"
#include "nfslib.h"
#include "misc.h"
#include "pseudoflavors.h"
#include "v4root.h"


#include <sys/utsname.h>

static inline unsigned int MAKE_VERSION(unsigned int p, unsigned int q,
					unsigned int r)
{
	return (65536 * p) + (256 * q) + r;
}

static inline unsigned int linux_version_code(void)
{
	struct utsname my_utsname;
	unsigned int p, q, r;

	if (uname(&my_utsname))
		return 0;

	p = atoi(strtok(my_utsname.release, "."));
	q = atoi(strtok(NULL, "."));
	r = atoi(strtok(NULL, "."));
	return MAKE_VERSION(p, q, r);
}

#ifndef _PATH_PSEUDO_ROOT
#define _PATH_PSEUDO_ROOT		"/"
#endif

#ifndef _PSEUDO_ROOT_FSID
#define _PSEUDO_ROOT_FSID	0
#endif

void v4root_set(void);
void v4root_unset(void);
static int v4root_support(void);

static struct exportent *v4root_create(char *, nfs_export *);

int v4root_needed;
static nfs_export pseudo_root = {
	.m_next = NULL,
	.m_client = NULL,
	.m_export = {
		.e_hostname = "*",
		.e_path = _PATH_PSEUDO_ROOT,
		.e_flags = NFSEXP_READONLY | NFSEXP_ROOTSQUASH
				| NFSEXP_NOSUBTREECHECK | NFSEXP_FSID
				| NFSEXP_CROSSMOUNT | NFSEXP_V4ROOT,
		.e_anonuid = 65534,
		.e_anongid = 65534,
		.e_squids = NULL,
		.e_nsquids = 0,
		.e_sqgids = NULL,
		.e_nsqgids = 0,
		.e_fsid = 0,
		.e_mountpoint = NULL,
		.e_secinfo[0].flav = NULL,
	},
	.m_exported = 0,
	.m_xtabent = 1,
	.m_mayexport = 1,
	.m_changed = 0,
	.m_warned = 0,
};

/*
 * Create a pseudo export
 */
static struct exportent *
v4root_create(char *path, nfs_export *export)
{
	nfs_export *exp;
	struct exportent eep;
	struct exportent *curexp = &export->m_export;

	dupexportent(&eep, curexp);
	eep.e_hostname = strdup(curexp->e_hostname);
	strncpy(eep.e_path, path, sizeof(eep.e_path));
	exp = export_create(&eep, 0);
	if (exp == NULL)
		return NULL;

	exp->m_export.e_flags |= NFSEXP_V4ROOT;
	if (strcmp(path, _PATH_PSEUDO_ROOT) != 0)
		exp->m_export.e_flags &= ~NFSEXP_FSID;

	xlog(D_CALL, "v4root_create: path '%s'", exp->m_export.e_path);

	return &exp->m_export;
}

/*
 * Make sure the kernel has pseudo root support.
 */
static int
v4root_support()
{
	static int kernel_support = -1;
	char *ptr, version[64];
	int major, minor;
	FILE *fp;

	if (kernel_support != -1)
		return kernel_support;

	if (linux_version_code() >= MAKE_VERSION(2, 6, 33)) {
		kernel_support = 1;
		return kernel_support;
	}

	kernel_support = 0;
	fp = fopen("/proc/fs/nfsd/exports", "r");
	if (fp == NULL)
		goto out;

	ptr = fgets(version, 64, fp);
	fclose(fp);
	if (ptr == NULL)
		goto out;

	while(*ptr && isdigit(*ptr) == 0)
		ptr++;
	if (*ptr == '\0')
		goto out;

	major = minor = 0;
	sscanf(ptr, " %d.%d",&major, &minor);
	if (major >= 1 && minor >= 2)
		kernel_support = 1;
out:
	if (!kernel_support) {
		xlog(L_WARNING, "Kernel does not have pseudo root support.");
		xlog(L_WARNING, "NFS v4 mounts will be disabled unless fsid=0");
		xlog(L_WARNING, "is specfied in /etc/exports file.");
	}

	return kernel_support;
}

/*
 * Create pseudo exports by running through the real export 
 * looking at the components of the path that make up the export. 
 * Those path components, if not exported, will become pseudo 
 * exports allowing them to be found when the kernel does an upcall 
 * looking for components of the v4 mount.
 */
void
v4root_set()
{
	nfs_export	*exp, *nxt;
	struct exportent *proot;
	int	i, insecure = 0, secflavors = 0;
	char *path, *ptr;
	char *hostname;

	if (!v4root_needed)
		return;

	if (!v4root_support())
		return;

	proot = v4root_create(_PATH_PSEUDO_ROOT, &pseudo_root);
	if (proot == NULL) {
		xlog(L_WARNING, "v4root_set: Unable to create "
			"pseudo export for '%s'", _PATH_PSEUDO_ROOT);
		return;
	}

	for (i = 0; i < MCL_MAXTYPES; i++) {
		for (exp = exportlist[i].p_head; exp; exp = nxt) {
			nxt = exp->m_next;
			hostname = exp->m_export.e_hostname;

			path = strdup(exp->m_export.e_path);
			ptr = path + 1;
			while ((ptr = strchr(ptr, '/')) != NULL) {
				*ptr = '\0';
				if (export_lookup(hostname, path, 0) == NULL)
					if (v4root_create(path, exp) == NULL) {
						xlog(L_WARNING, "v4root_set: Unable to create "
							"pseudo export for '%s'", path);
						break;
					}
				*ptr = '/';
				ptr++;
			}
			/* Make note of insecure exports */
			if (!insecure)
				insecure = (exp->m_export.e_flags & NFSEXP_INSECURE_PORT);

			/* Make note of security flavors being set */
			if (!secflavors)
				secflavors = (exp->m_export.e_secinfo[0].flav != NULL);

			free(path);
		}
	}

	/*
	 * If there are any insecure exports, the pseudo root
	 * also has to be insecure
	 */
	if (insecure) {
		proot->e_flags |= NFSEXP_INSECURE_PORT;
	}

	/*
	 * Make sure the pseudo root is accessible from all
	 * security flavors when it needs to be.
	 */
	if (secflavors && !flavors_setall(proot)) {
		xlog(L_WARNING, "v4root_set: Unable to set security "
			"flavors on pseudo root");
	}

}

